@@ -18,11 +18,16 @@
 #define TILEDB_DEPRECATED
 #define TILEDB_DEPRECATED_EXPORT
 
-#include <tiledb/arrowio>
 #include <tiledb/tiledb.h>              // C
 #include <tiledb/tiledb>                // C++
 #include <tiledb/tiledb_experimental.h> // C
 
+// clang-format off
+// do not re-order these headers
+#include "py_arrowio"
+#include "py_arrow_io_impl.h"
+// clang-format on
+
 #if defined(TILEDB_SERIALIZATION)
 #include <tiledb/tiledb_serialization.h> // C
 #endif
@@ -120,6 +125,17 @@ struct BufferInfo {
   py::array_t<uint8_t> validity;
 };
 
+struct BufferHolder {
+  py::object data;
+  py::object offsets;
+  py::object validity;
+
+  BufferHolder(py::object d, py::object o, py::object v)
+      : data(d), offsets(o), validity(v) {}
+
+  static void free_buffer_holder(BufferHolder *self) { delete self; }
+};
+
 py::dtype tiledb_dtype(tiledb_datatype_t type, uint32_t cell_val_num) {
   if (cell_val_num == 1) {
     auto np = py::module::import("numpy");
@@ -1348,8 +1364,8 @@ public:
 
     std::unique_ptr<PAPair> pa_pair(new PAPair());
 
-    adapter.export_buffer(name.c_str(), &(pa_pair->array_),
-                          &(pa_pair->schema_));
+    adapter.export_buffer(name.c_str(), &(pa_pair->array_), &(pa_pair->schema_),
+                          nullptr, nullptr);
     pa_pair->exported_ = true;
 
     return pa_pair;
@@ -1370,14 +1386,21 @@ public:
     py::list names;
     py::list results;
     for (auto &buffer_name : buffers_order_) {
+      BufferInfo &buffer_info = buffers_.at(buffer_name);
+
+      auto buffer_holder = new BufferHolder(
+          buffer_info.data, buffer_info.validity, buffer_info.offsets);
+
       ArrowArray c_pa_array;
       ArrowSchema c_pa_schema;
       adapter.export_buffer(buffer_name.c_str(),
                             static_cast<void *>(&c_pa_array),
-                            static_cast<void *>(&c_pa_schema));
+                            static_cast<void *>(&c_pa_schema),
+                            (tiledb::arrow::ArrowAdapter::release_cb)
+                                BufferHolder::free_buffer_holder,
+                            buffer_holder);
 
       if (is_nullable(buffer_name)) {
-        BufferInfo &buffer_info = buffers_.at(buffer_name);
         // count zeros before converting to bitmap
         c_pa_array.null_count = count_zeros(buffer_info.validity);
         // convert to bitmap
